from __future__ import annotations

from dataclasses import asdict
from typing import TYPE_CHECKING

from avilla.core.elements import Audio, Face, File, Notice, NoticeAll, Picture, Text, Video
from avilla.core.resource import LocalFileResource, RawResource, UrlResource
from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.qqapi.capability import QQAPICapability
from avilla.qqapi.element import Ark, Embed, Keyboard, Markdown, Reference
from avilla.qqapi.resource import (
    QQAPIAudioResource,
    QQAPIImageResource,
    QQAPIVideoResource,
    QQAPIFileResource,
)
from avilla.qqapi.utils import escape

if TYPE_CHECKING:
    from avilla.qqapi.account import QQAPIAccount  # noqa
    from avilla.qqapi.protocol import QQAPIProtocol  # noqa


class QQAPIMessageSerializePerform((m := AccountCollector["QQAPIProtocol", "QQAPIAccount"]())._):
    m.namespace = "avilla.protocol/qqapi::message"
    m.identify = "serialize"

    # LINK: https://github.com/microsoft/pyright/issues/5409

    @m.entity(QQAPICapability.serialize_element, element=Text)
    async def text(self, element: Text):
        return escape(element.text)

    @m.entity(QQAPICapability.serialize_element, element=Face)
    async def face(self, element: Face):
        return f"<emoji:{element.id}>"

    @m.entity(QQAPICapability.serialize_element, element=Notice)
    async def notice(self, element: Notice):
        if element.target.last_key == "channel":
            return f"<#{element.target['channel']}>"
        return f"<@{element.target['member']}>"

    @m.entity(QQAPICapability.serialize_element, element=NoticeAll)
    async def notice_all(self, element: NoticeAll):
        return "@everyone"

    @m.entity(QQAPICapability.serialize_element, element=Picture)
    async def picture(self, element: Picture):
        if isinstance(element.resource, (QQAPIImageResource, UrlResource)):
            return "media", ("image", element.resource.url)
        if isinstance(element.resource, LocalFileResource):
            return "file_image", element.resource.file.read_bytes()
        if isinstance(element.resource, RawResource):
            return "file_image", element.resource.data
        return "file_image", await self.account.staff.fetch_resource(element.resource)

    @m.entity(QQAPICapability.serialize_element, element=Audio)
    async def audio(self, element: Audio):
        if isinstance(element.resource, (QQAPIAudioResource, UrlResource)):
            return "media", ("audio", element.resource.url)
        if isinstance(element.resource, LocalFileResource):
            return "file_audio", element.resource.file.read_bytes()
        if isinstance(element.resource, RawResource):
            return "file_audio", element.resource.data
        return "file_audio", await self.account.staff.fetch_resource(element.resource)

    @m.entity(QQAPICapability.serialize_element, element=Video)
    async def video(self, element: Video):
        if isinstance(element.resource, (QQAPIVideoResource, UrlResource)):
            return "media", ("video", element.resource.url)
        if isinstance(element.resource, LocalFileResource):
            return "file_video", element.resource.file.read_bytes()
        if isinstance(element.resource, RawResource):
            return "file_video", element.resource.data
        return "file_video", await self.account.staff.fetch_resource(element.resource)

    @m.entity(QQAPICapability.serialize_element, element=File)
    async def file(self, element: File):
        if isinstance(element.resource, (QQAPIFileResource, UrlResource)):
            return "media", ("file", element.resource.url)
        if isinstance(element.resource, LocalFileResource):
            return "file_file", element.resource.file.read_bytes()
        if isinstance(element.resource, RawResource):
            return "file_file", element.resource.data
        return "file_file", await self.account.staff.fetch_resource(element.resource)

    @m.entity(QQAPICapability.serialize_element, element=Reference)
    async def reference(self, element: Reference):
        return "message_reference", {
            "message_id": element.message["message"],
            "ignore_get_message_error": element.ignore_get_message_error,
        }

    @m.entity(QQAPICapability.serialize_element, element=Embed)
    async def embed(self, element: Embed):
        res = {
            "title": element.title,
            "prompt": element.prompt,
            "thumbnail": {"url": element.thumbnail} if element.thumbnail else None,
            "fields": [{"name": i} for i in element.fields] if element.fields else None,
        }
        return "embed", {k: v for k, v in res.items() if v}

    @m.entity(QQAPICapability.serialize_element, element=Ark)
    async def ark(self, element: Ark):
        return "ark", {
            "template_id": element.template_id,
            "kv": [
                (
                    {
                        "key": kv.key,
                        "obj": [{"obj_kv": [{"key": k, "value": v} for k, v in obj_kv.items()]} for obj_kv in kv.obj],
                    }
                    if kv.obj
                    else {"key": kv.key, "value": kv.value}
                )
                for kv in element.kv
            ]
            if element.kv
            else None,
        }

    @m.entity(QQAPICapability.serialize_element, element=Markdown)
    async def markdown(self, element: Markdown):
        if element.params:
            param = [{"key": k, "values": v} for k, v in element.params.items()]
        else:
            param = None
        return "markdown", {
            "content": element.content,
            "custom_template_id": element.custom_template_id,
            "params": param,
        }

    @m.entity(QQAPICapability.serialize_element, element=Keyboard)
    async def keyboard(self, element: Keyboard):
        content = {"rows": []}
        for row in element.content or []:
            buttons = {"buttons": []}
            for button in row:
                raw = asdict(button)
                raw["render_data"] = asdict(button.render_data)
                raw["action"] = asdict(button.action)
                raw["action"]["permission"] = asdict(button.action.permission)
                buttons["buttons"].append(raw)
            content["rows"].append(buttons)

        return "keyboard", {"id": element.id, "content": content}



import json
import re


def escape(s: str) -> str:
    return s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")


def unescape(s: str) -> str:
    return s.replace("&lt;", "<").replace("&gt;", ">").replace("&amp;", "&")


def handle_text(msg: str):
    text_begin = 0
    msg = msg.replace("@everyone", "")
    for embed in re.finditer(
        r"\<(?P<type>(?:@|#|emoji:))!?(?P<id>\w+?)\>",
        msg,
    ):
        if content := msg[text_begin : embed.pos + embed.start()]:
            yield {"type": "text", "text": unescape(content)}
        text_begin = embed.pos + embed.end()
        if embed["type"] == "@":
            yield {"type": "mention_user", "user_id": embed.group("id")}
        elif embed["type"] == "#":
            yield {"type": "mention_channel", "channel_id": embed.group("id")}
        else:
            yield {"type": "emoji", "id": embed.group("id")}
    if content := msg[text_begin:]:
        yield {"type": "text", "text": unescape(content)}


def form_data(message: dict):
    if not (file_image := message.pop("file_image", None)):
        return "post", message
    files = {"file_image": {"value": file_image, "content_type": None, "filename": "file_image"}}
    data_ = {}
    for key, value in message.items():
        if isinstance(value, (list, dict)):
            files[key] = {
                "value": json.dumps({key: value}).encode("utf-8"),
                "content_type": "application/json",
                "filename": f"{key}.json",
            }
        else:
            data_[key] = value
    return "multipart", {"files": files, "data": data_}



from __future__ import annotations

import asyncio
import json
import sys
from contextlib import suppress
from dataclasses import asdict
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, cast

import aiohttp
from launart import Service
from launart.manager import Launart
from launart.utilles import any_completed
from loguru import logger

from avilla.core.account import AccountInfo
from avilla.core.selector import Selector
from avilla.qqapi.account import QQAPIAccount
from avilla.qqapi.const import PLATFORM
from avilla.qqapi.exception import NetworkError, UnauthorizedException
from avilla.standard.core.account import (
    AccountAvailable,
    AccountRegistered,
    AccountUnavailable,
    AccountUnregistered,
)

from .base import CallMethod, QQAPINetworking
from .util import Opcode, Payload, validate_response

if TYPE_CHECKING:
    from avilla.qqapi.protocol import QQAPIConfig, QQAPIProtocol


class QQAPIWsClientNetworking(QQAPINetworking, Service):
    required: set[str] = set()
    stages: set[str] = {"preparing", "blocking", "cleanup"}

    config: QQAPIConfig
    connections: dict[tuple[int, int], aiohttp.ClientWebSocketResponse]
    session: aiohttp.ClientSession

    @property
    def id(self):
        return f"qqapi/connection/client#{self.config.id}"

    def __init__(self, protocol: QQAPIProtocol, config: QQAPIConfig) -> None:
        super().__init__(protocol)
        self.config = config
        if any([not config.id, not config.token, not config.secret]):
            raise ValueError("config is not complete")
        self.connections = {}

    async def get_access_token(self) -> str:
        if self._access_token is None or (
            self._expires_in and datetime.now(timezone.utc) > self._expires_in - timedelta(seconds=30)
        ):
            async with self.session.post(
                self.config.auth_base,
                json={
                    "appId": self.config.id,
                    "clientSecret": self.config.secret,
                },
            ) as resp:
                if resp.status != 200 or not resp.content:
                    raise NetworkError(
                        f"Get authorization failed with status code {resp.status}." " Please check your config."
                    )
                data = await resp.json()
            self._access_token = cast(str, data["access_token"])
            self._expires_in = datetime.now(timezone.utc) + timedelta(seconds=int(data["expires_in"]))
        return self._access_token

    async def _get_authorization_header(self) -> str:
        """获取当前 Bot 的鉴权信息"""
        if self.config.is_group_bot:
            return f"QQBot {await self.get_access_token()}"
        return f"Bot {self.config.id}.{self.config.token}"

    async def get_authorization_header(self) -> dict[str, str]:
        """获取当前 Bot 的鉴权信息"""
        headers = {"Authorization": await self._get_authorization_header()}
        if self.config.is_group_bot:
            headers["X-Union-Appid"] = self.config.id
        return headers

    async def message_receive(self, shard: tuple[int, int]):
        if (connection := self.connections.get(shard)) is None:
            raise RuntimeError("connection is not established")

        async for msg in connection:
            # logger.debug(f"{msg=}")

            if msg.type in {aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.ERROR, aiohttp.WSMsgType.CLOSED}:
                self.close_signal.set()
                break
            elif msg.type == aiohttp.WSMsgType.TEXT:
                data: dict = json.loads(cast(str, msg.data))
                if data["op"] == Opcode.RECONNECT:
                    logger.warning("Received reconnect event from server, will reconnect in 5 seconds...")
                    break
                if data["op"] == Opcode.INVALID_SESSION:
                    self.session_id = None
                    self.sequence = None
                    logger.warning("Received invalid session event from server, will try to resume")
                    break
                if data["op"] == Opcode.HEARTBEAT_ACK:
                    continue
                yield self, data
        else:
            await self.connection_closed()

    async def connection_closed(self):
        self.session_id = None
        self.sequence = None
        self.close_signal.set()

    async def send(self, payload: dict, shard: tuple[int, int]):
        if (connection := self.connections.get(shard)) is None:
            raise RuntimeError("connection is not established")

        await connection.send_json(payload)

    async def _call_http(
        self, method: CallMethod, action: str, headers: dict[str, str] | None = None, params: dict | None = None
    ) -> dict:
        params = params or {}
        params = {k: v for k, v in params.items() if v is not None}
        if method in {"get", "fetch"}:
            async with self.session.get(
                (self.config.get_api_base() / action).with_query(params),
                headers=headers,
            ) as resp:
                return await validate_response(resp)

        if method == "patch":
            async with self.session.patch(
                (self.config.get_api_base() / action),
                json=params,
                headers=headers,
            ) as resp:
                return await validate_response(resp)

        if method == "put":
            async with self.session.put(
                (self.config.get_api_base() / action),
                json=params,
                headers=headers,
            ) as resp:
                return await validate_response(resp)

        if method == "delete":
            async with self.session.delete(
                (self.config.get_api_base() / action).with_query(params),
                headers=headers,
            ) as resp:
                return await validate_response(resp)

        if method in {"post", "update"}:
            async with self.session.post(
                (self.config.get_api_base() / action),
                json=params,
                headers=headers,
            ) as resp:
                return await validate_response(resp)

        if method == "multipart":
            if params is None:
                raise TypeError("multipart requires params")
            data = aiohttp.FormData(params["data"], quote_fields=False)
            for k, v in params["files"].items():
                if isinstance(v, dict):
                    data.add_field(k, v["value"], filename=v.get("filename"), content_type=v.get("content_type"))
                else:
                    data.add_field(k, v)

            async with self.session.post(
                (self.config.get_api_base() / action),
                data=data,
                headers=headers,
            ) as resp:
                return await validate_response(resp)

        raise ValueError(f"unknown method {method}")

    async def call_http(self, method: CallMethod, action: str, params: dict | None = None) -> dict:
        headers = await self.get_authorization_header()
        try:
            return await self._call_http(method, action, headers, params)
        except UnauthorizedException as e:
            if not self.config.is_group_bot:
                raise
            self._access_token = None
            try:
                headers = await self.get_authorization_header()
            except Exception:
                raise e from None
            try:
                return await self._call_http(method, action, headers, params)
            except Exception as e1:
                raise e1 from None

    async def wait_for_available(self):
        await self.status.wait_for_available()

    def get_staff_components(self):
        return {"connection": self, "protocol": self.protocol, "avilla": self.protocol.avilla}

    def get_staff_artifacts(self):
        return [self.protocol.artifacts, self.protocol.avilla.global_artifacts]

    def __staff_generic__(self, element_type: dict, event_type: dict):
        ...

    @property
    def alive(self):
        return self.connections and any(not connection.closed for connection in self.connections.values())

    async def _hello(self, shard: tuple[int, int]) -> int | None:
        """接收并处理服务器的 Hello 事件"""
        if not (connection := self.connections.get(shard)):
            raise RuntimeError("connection is not established")
        try:
            payload = Payload(**await connection.receive_json())
            assert payload.opcode == Opcode.HELLO, f"Received unexpected payload: {payload!r}"
            return payload.data["heartbeat_interval"]
        except Exception as e:
            logger.error(
                "Error while receiving server hello event",
                e,
            )

    async def _authenticate(self, shard: tuple[int, int]):
        """鉴权连接"""
        if not (connection := self.connections.get(shard)):
            raise RuntimeError("connection is not established")
        if not self.session_id:
            payload = Payload(
                op=Opcode.IDENTIFY,
                d={
                    "token": await self._get_authorization_header(),
                    "intents": self.config.intent.to_int(),
                    "shard": list(shard),
                    "properties": {
                        "$os": sys.platform,
                        "$language": f"python {sys.version}",
                        "$sdk": "Avilla",
                    },
                },
            )
        else:
            payload = Payload(
                op=Opcode.RESUME,
                d={
                    "token": await self._get_authorization_header(),
                    "session_id": self.session_id,
                    "seq": self.sequence,
                },
            )

        try:
            await self.send(asdict(payload), shard)
        except Exception as e:
            logger.error(f"Error while sending {payload.opcode.name.title()} event: {e}")
            return False

        if not self.session_id:
            # https://bot.q.qq.com/wiki/develop/api/gateway/reference.html#_2-%E9%89%B4%E6%9D%83%E8%BF%9E%E6%8E%A5
            # 鉴权成功之后，后台会下发一个 Ready Event
            payload = Payload(**await connection.receive_json())
            if payload.opcode == Opcode.INVALID_SESSION:
                logger.warning("Received invalid session event from server, will try to resume")
                return False
            if not (payload.opcode == Opcode.DISPATCH and payload.type == "READY" and payload.data):
                logger.error(f"Received unexpected payload: {payload}")
                return False
            self.sequence = payload.sequence
            self.session_id = payload.data["session_id"]
            self.self_info = payload.data["user"]
            self.account_id = payload.data["user"]["id"]
            account_route = Selector().land("qq").account(self.account_id)
            if account_route in self.protocol.avilla.accounts:
                account = cast(QQAPIAccount, self.protocol.avilla.accounts[account_route].account)
            else:
                account = QQAPIAccount(account_route, self.protocol)
                self.protocol.avilla.accounts[account_route] = AccountInfo(
                    account_route,
                    account,
                    self.protocol,
                    PLATFORM,
                )
            self.protocol.service.accounts[self.account_id] = account
            account.connection = self
            self.protocol.avilla.broadcast.postEvent(AccountRegistered(self.protocol.avilla, account))
        else:
            account = self.protocol.service.accounts[self.account_id]
            account.connection = self
        self.protocol.avilla.broadcast.postEvent(AccountAvailable(self.protocol.avilla, account))
        return True

    async def _heartbeat(self, heartbeat_interval: int, shard: tuple[int, int]):
        """心跳"""
        while True:
            if self.session_id:
                with suppress(Exception):
                    await self.send({"op": 1, "d": self.sequence}, shard=shard)
            await asyncio.sleep(heartbeat_interval / 1000)

    async def connection_daemon(
        self, manager: Launart, session: aiohttp.ClientSession, url: str, shard: tuple[int, int]
    ):
        while not manager.status.exiting:
            try:
                async with session.ws_connect(url, timeout=30) as conn:
                    self.connections[shard] = conn
                    logger.info(f"{self.id} Websocket client connected")
                    heartbeat_interval = await self._hello(shard)
                    if not heartbeat_interval:
                        await asyncio.sleep(3)
                        continue
                    result = await self._authenticate(shard)
                    if not result:
                        await asyncio.sleep(3)
                        continue
                    account_route = Selector().land("qq").account(self.account_id)
                    self.close_signal.clear()
                    close_task = asyncio.create_task(self.close_signal.wait())
                    receiver_task = asyncio.create_task(self.message_handle(shard))
                    sigexit_task = asyncio.create_task(manager.status.wait_for_sigexit())
                    heartbeat_task = asyncio.create_task(self._heartbeat(heartbeat_interval, shard))
                    done, pending = await any_completed(
                        sigexit_task,
                        close_task,
                        receiver_task,
                        heartbeat_task,
                    )
                    if sigexit_task in done:
                        logger.info(f"{self} Websocket client exiting...")
                        await conn.close()
                        self.close_signal.set()
                        receiver_task.cancel()
                        heartbeat_task.cancel()
                        with suppress(KeyError):
                            del self.connections[shard]
                            await self.protocol.avilla.broadcast.postEvent(
                                AccountUnregistered(
                                    self.protocol.avilla, self.protocol.avilla.accounts[account_route].account
                                )
                            )
                            del self.protocol.service.accounts[self.account_id]
                            del self.protocol.avilla.accounts[account_route]
                        return
                    if close_task in done:
                        receiver_task.cancel()
                        heartbeat_task.cancel()
                        logger.warning(f"{self} Connection closed by server, will reconnect in 5 seconds...")

                        with suppress(KeyError):
                            await self.protocol.avilla.broadcast.postEvent(
                                AccountUnavailable(
                                    self.protocol.avilla, self.protocol.avilla.accounts[account_route].account
                                )
                            )
                            del self.protocol.service.accounts[self.account_id]
                            # del self.protocol.avilla.accounts[account_route]
                        await asyncio.sleep(5)
                        logger.info(f"{self} Reconnecting...")
                        continue
            except Exception as e:
                logger.error(f"{self} Error while connecting: {e}")
                await asyncio.sleep(5)
                logger.info(f"{self} Reconnecting...")

    async def launch(self, manager: Launart):
        async with self.stage("preparing"):
            self.session = aiohttp.ClientSession()
            gateway_info = await self.call_http("get", "gateway/bot")
            ws_url = gateway_info["url"]
            remain = gateway_info.get("session_start_limit", {}).get("remaining")
            if remain is not None and remain <= 0:
                logger.error("Session start limit reached, please wait for a while")
                manager.status.exiting = True
                return
        tasks = []
        async with self.stage("blocking"):
            if self.config.shard:
                tasks.append(
                    asyncio.create_task(self.connection_daemon(manager, self.session, ws_url, self.config.shard))
                )
            else:
                shards = gateway_info.get("shards") or 1
                logger.debug(f"Get Shards: {shards}")
                for i in range(shards):
                    tasks.append(
                        asyncio.create_task(self.connection_daemon(manager, self.session, ws_url, (i, shards)))
                    )
                    await asyncio.sleep(gateway_info.get("session_start_limit", {}).get("max_concurrency", 1))
            await any_completed(*tasks)

        async with self.stage("cleanup"):
            await self.session.close()
            for task in tasks:
                task.cancel()
            await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
            self.connections.clear()



from dataclasses import dataclass

from avilla.core.metadata import Metadata
from avilla.core.selector import Selector


@dataclass
class Role(Metadata):
    id: int
    scene: Selector
    name: str
    hoist: bool
    color: int

    def to_selector(self):
        return self.scene.role(str(self.id))



from __future__ import annotations

from typing import TYPE_CHECKING

from avilla.core.elements import (
    Audio,
    Face,
    File,
    Notice,
    NoticeAll,
    Picture,
    Text,
    Video,
)
from avilla.core.ryanvk.collector.application import ApplicationCollector
from avilla.core.selector import Selector
from avilla.qqapi.capability import QQAPICapability
from avilla.qqapi.element import Ark, ArkKv, Embed, Reference
from avilla.qqapi.resource import (
    QQAPIAudioResource,
    QQAPIFileResource,
    QQAPIImageResource,
    QQAPIVideoResource,
)
from graia.ryanvk import OptionalAccess

if TYPE_CHECKING:
    from avilla.core.context import Context


class QQAPIMessageDeserializePerform((m := ApplicationCollector())._):
    m.namespace = "avilla.protocol/qqapi::message"
    m.identify = "deserialize"

    # LINK: https://github.com/microsoft/pyright/issues/5409
    context: OptionalAccess[Context] = OptionalAccess()

    @m.entity(QQAPICapability.deserialize_element, raw_element="text")
    async def text(self, raw_element: dict) -> Text:
        return Text(raw_element["text"])

    @m.entity(QQAPICapability.deserialize_element, raw_element="emoji")
    async def emoji(self, raw_element: dict) -> Face:
        return Face(raw_element["id"])

    @m.entity(QQAPICapability.deserialize_element, raw_element="attachment")
    async def attachment(self, raw_element: dict):
        if "content_type" not in raw_element:
            resource = QQAPIImageResource(Selector().land("qqguild").picture(url := raw_element["url"]), "image", url)
            return Picture(resource)
        content_type = raw_element["content_type"].split("/")[0]
        if content_type == "file":
            resource = QQAPIFileResource(
                Selector().land("qqguild").file(url := raw_element["url"]),
                raw_element["content_type"],
                url,
                raw_element.get("filename"),
                raw_element.get("height"),
                raw_element.get("width"),
                raw_element.get("size"),
            )
            return File(resource)
        if content_type == "audio":
            resource = QQAPIAudioResource(
                Selector().land("qqguild").audio(url := raw_element["url"]),
                raw_element["content_type"],
                url,
                raw_element.get("filename"),
                raw_element.get("height"),
                raw_element.get("width"),
                raw_element.get("size"),
            )
            return Audio(resource)
        if content_type == "video":
            resource = QQAPIVideoResource(
                Selector().land("qqguild").video(url := raw_element["url"]),
                raw_element["content_type"],
                url,
                raw_element.get("filename"),
                raw_element.get("height"),
                raw_element.get("width"),
                raw_element.get("size"),
            )
            return Video(resource)
        resource = QQAPIImageResource(
            Selector().land("qqguild").picture(url := raw_element["url"]),
            raw_element["content_type"],
            url,
            raw_element.get("filename"),
            raw_element.get("height"),
            raw_element.get("width"),
            raw_element.get("size"),
        )
        return Picture(resource)

    @m.entity(QQAPICapability.deserialize_element, raw_element="mention_user")
    async def mention(self, raw_element: dict) -> Notice:
        if self.context:
            return Notice(self.context.scene.member(raw_element["user_id"]))
        return Notice(Selector().land("qqguild").member(raw_element["user_id"]))

    @m.entity(QQAPICapability.deserialize_element, raw_element="mention_channel")
    async def mention_channel(self, raw_element: dict) -> Notice:
        if self.context:
            return Notice(
                Selector().land("qqguild").guild(self.context.scene["guild"]).channel(raw_element["channel_id"])
            )
        return Notice(Selector().land("qqguild").channel(raw_element["channel_id"]))

    @m.entity(QQAPICapability.deserialize_element, raw_element="mention_everyone")
    async def mention_everyone(self, raw_element: dict) -> NoticeAll:
        return NoticeAll()

    @m.entity(QQAPICapability.deserialize_element, raw_element="message_reference")
    async def message_reference(self, raw_element: dict) -> Reference:
        if self.context:
            return Reference(self.context.scene.message(raw_element["message_id"]))
        return Reference(Selector().land("qqguild").message(raw_element["message_id"]))

    @m.entity(QQAPICapability.deserialize_element, raw_element="embed")
    async def embed(self, raw_element: dict) -> Embed:
        return Embed(
            raw_element["title"],
            raw_element.get("prompt"),
            raw_element.get("thumbnail", {})["url"],
            [i["name"] for i in raw_element.get("fields", [])],
        )

    @m.entity(QQAPICapability.deserialize_element, raw_element="ark")
    async def ark(self, raw_element: dict) -> Ark:
        kvs: list[ArkKv] = []
        for i in raw_element.get("kv", []):
            if i.get("value"):
                kvs.append(ArkKv(**i))
            else:
                obj: list[dict] = [{slot["key"]: slot["value"] for slot in kv["obj_kv"]} for kv in i["obj"]]
                kvs.append(ArkKv(i["key"], obj=obj))
        return Ark(raw_element["template_id"], kvs)



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING

from avilla.core.exceptions import permission_error_message
from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.core.selector import Selector
from avilla.qqapi.const import PRIVILEGE_TRANS
from avilla.standard.core.privilege import MuteCapability, MuteInfo, Privilege
from avilla.standard.core.profile import Avatar, Nick, Summary
from avilla.standard.core.relation import SceneCapability

if TYPE_CHECKING:
    from avilla.qqapi.account import QQAPIAccount  # noqa
    from avilla.qqapi.protocol import QQAPIProtocol  # noqa


class QQAPIGuildMemberActionPerform((m := AccountCollector["QQAPIProtocol", "QQAPIAccount"]())._):
    m.namespace = "avilla.protocol/qqapi::action"
    m.identify = "guild_member"

    @m.pull("land.guild.member", Nick)
    @m.pull("land.guild.channel.member", Nick)
    async def get_nick(self, target: Selector, route: ...) -> Nick:
        result = await self.account.connection.call_http(
            "get", f"guilds/{target.pattern['guild']}/members/{target.pattern['member']}", {}
        )
        return Nick(result["user"]["username"], result["nick"], None)

    @m.pull("land.guild.member", Summary)
    @m.pull("land.guild.channel.member", Summary)
    async def get_summary(self, target: Selector, route: ...) -> Summary:
        result = await self.account.connection.call_http(
            "get", f"guilds/{target.pattern['guild']}/members/{target.pattern['member']}", {}
        )
        return Summary(result["user"]["username"], result["nick"])

    @m.pull("land.guild.member", Avatar)
    @m.pull("land.guild.channel.member", Avatar)
    async def get_avatar(self, target: Selector, route: ...) -> Avatar:
        result = await self.account.connection.call_http(
            "get", f"guilds/{target.pattern['guild']}/members/{target.pattern['member']}", {}
        )
        return Avatar(result["user"]["avatar"])

    @m.pull("land.guild.channel.member", Privilege)
    async def get_privilege(self, target: Selector, route: ...) -> Privilege:
        self_info = await self.account.connection.call_http(
            "get", f"guilds/{target.pattern['guild']}/members/{self.account.route['account']}", {}
        )
        target_info = await self.account.connection.call_http(
            "get", f"guilds/{target.pattern['guild']}/members/{target.pattern['member']}", {}
        )
        effective = False
        if "4" in self_info["roles"]:
            effective = True
        elif "2" in self_info["roles"] and "4" not in target_info["roles"] and "2" not in target_info["roles"]:
            effective = True
        elif "5" in self_info["roles"] and target_info["roles"] == ["1"]:
            effective = True
        apis = await self.account.connection.call_http("get", f"guilds/{target.pattern['guild']}/api_permission", {})
        for api in apis["apis"]:
            if api["path"] == "/channels/{channel_id}/members/{user_id}/permissions" and api["method"] == "GET":
                self_permission = await self.account.connection.call_http(
                    "get",
                    f"channels/{target.pattern['channel']}/members/{self.account.route['account']}/permissions",
                    {},
                )
                result = await self.account.connection.call_http(
                    "get", f"channels/{target.pattern['channel']}/members/{target.pattern['member']}/permissions", {}
                )
                return Privilege(
                    int(result["permissions"]) & 2 == 2,
                    effective or (int(self_permission["permissions"]) & 2 == 2),
                )
        raise PermissionError(permission_error_message(f"get_permission@{target.path}", "read", ["manage"]))

    @m.pull("land.guild.channel.member", Privilege >> Summary)
    async def get_privilege_summary(self, target: Selector, route: ...) -> Summary:
        apis = await self.account.connection.call_http("get", f"guilds/{target.pattern['guild']}/api_permission", {})
        for api in apis["apis"]:
            if api["path"] == "/channels/{channel_id}/members/{user_id}/permissions" and api["method"] == "GET":
                result = await self.account.connection.call_http(
                    "get", f"channels/{target.pattern['channel']}/members/{target.pattern['member']}/permissions", {}
                )
                return Summary(
                    PRIVILEGE_TRANS[int(result["permissions"])],
                    "user's permissions in channel",
                )
        raise PermissionError(permission_error_message(f"get_permission@{target.path}", "read", ["manage"]))

    @m.pull("land.guild.channel.member", MuteInfo)
    async def get_mute_info(self, target: Selector, route: ...) -> MuteInfo:
        apis = await self.account.connection.call_http("get", f"guilds/{target.pattern['guild']}/api_permission", {})
        for api in apis["apis"]:
            if api["path"] == "/channels/{channel_id}/members/{user_id}/permissions" and api["method"] == "GET":
                result = await self.account.connection.call_http(
                    "get", f"channels/{target.pattern['channel']}/members/{target.pattern['member']}/permissions", {}
                )
                return MuteInfo(
                    int(result["permissions"]) & 4 == 4,
                    timedelta(seconds=0),
                )
        raise PermissionError(permission_error_message(f"get_permission@{target.path}", "read", ["manage"]))

    @MuteCapability.mute.collect(m, target="land.guild.channel.member")
    async def mute(self, target: Selector, duration: timedelta) -> None:
        if not (await self.get_privilege(target, Privilege)).effective:
            self_info = await self.get_privilege_summary(
                target.into(f"~.member{self.account.route['account']}"), Summary
            )
            raise PermissionError(permission_error_message(f"set_permission@{target.path}", self_info.name, ["manage"]))
        await self.account.connection.call_http(
            "put",
            f"channels/{target.pattern['channel']}/members/{target.pattern['member']}/permissions",
            {"add": 0, "remove": 4},
        )

    @MuteCapability.unmute.collect(m, target="land.guild.channel.member")
    async def unmute(self, target: Selector) -> None:
        if not (await self.get_privilege(target, Privilege)).effective:
            self_info = await self.get_privilege_summary(
                target.into(f"~.member{self.account.route['account']}"), Summary
            )
            raise PermissionError(permission_error_message(f"set_permission@{target.path}", self_info.name, ["manage"]))
        await self.account.connection.call_http(
            "put",
            f"channels/{target.pattern['channel']}/members/{target.pattern['member']}/permissions",
            {"add": 4, "remove": 0},
        )

    @SceneCapability.remove_member.collect(m, target="land.guild.member")
    @SceneCapability.remove_member.collect(m, target="land.guild.channel.member")
    async def remove_user(self, target: Selector, reason: str | None = None) -> None:
        self_info = await self.account.connection.call_http(
            "get", f"guilds/{target.pattern['guild']}/members/{self.account.route['account']}", {}
        )
        effective = bool({"2", "4", "5"} & set(self_info["roles"]))
        if not effective:
            raise PermissionError(permission_error_message(f"remove_member@{target.path}", "read", ["manage"]))
        await self.account.connection.call_http(
            "delete", f"guilds/{target.pattern['guild']}/members/{target.pattern['member']}", {}
        )

    @SceneCapability.remove_member.collect(m, target="land.guild.channel.member")
    async def remove_member(self, target: Selector, reason: str | None = None) -> None:
        self_info = await self.account.connection.call_http(
            "get", f"guilds/{target.pattern['guild']}/members/{self.account.route['account']}", {}
        )
        effective = bool({"2", "4", "5"} & set(self_info["roles"]))
        if not effective:
            raise PermissionError(permission_error_message(f"remove_member@{target.path}", "read", ["manage"]))
        await self.account.connection.call_http(
            "delete", f"guilds/{target.pattern['guild']}/members/{target.pattern['member']}", {}
        )



from __future__ import annotations

from dataclasses import dataclass, field

from loguru import logger
from yarl import URL

from avilla.core.application import Avilla
from avilla.core.protocol import BaseProtocol, ProtocolConfig
from graia.ryanvk import merge, ref

from .connection.ws_client import QQAPIWsClientNetworking
from .service import QQAPIService


@dataclass
class Intents:
    guilds: bool = True
    guild_members: bool = True
    guild_messages: bool = False
    """GUILD_MESSAGES"""
    guild_message_reactions: bool = True
    direct_message: bool = False
    """DIRECT_MESSAGES"""
    open_forum_event: bool = False
    audio_live_member: bool = False
    c2c_group_at_messages: bool = False
    interaction: bool = False
    message_audit: bool = True
    forum_event: bool = False
    audio_action: bool = False
    at_messages: bool = True
    """PUBLIC_GUILD_MESSAGES"""

    def __post_init__(self):
        if self.at_messages and self.guild_messages:
            logger.warning("at_messages and guild_messages are both enabled, which is not recommended.")

    def to_int(self) -> int:
        return (
            self.guilds << 0
            | self.guild_members << 1
            | self.guild_messages << 9
            | self.guild_message_reactions << 10
            | self.direct_message << 12
            | self.open_forum_event << 18
            | self.audio_live_member << 19
            | self.c2c_group_at_messages << 25
            | self.interaction << 26
            | self.message_audit << 27
            | self.forum_event << 28
            | self.audio_action << 29
            | self.at_messages << 30
        )

    @property
    def is_group_enabled(self) -> bool:
        """是否开启群聊功能"""
        return self.c2c_group_at_messages is True


@dataclass
class QQAPIConfig(ProtocolConfig):
    id: str
    token: str
    secret: str
    shard: tuple[int, int] | None = None
    intent: Intents = field(default_factory=Intents)
    is_sandbox: bool = False
    api_base: URL = URL("https://api.sgroup.qq.com/")
    sandbox_api_base: URL = URL("https://sandbox.api.sgroup.qq.com")
    auth_base: URL = URL("https://bots.qq.com/app/getAppAccessToken")

    def get_api_base(self) -> URL:
        return URL(self.sandbox_api_base) if self.is_sandbox else URL(self.api_base)

    @property
    def is_group_bot(self) -> bool:
        """是否为群机器人"""
        return self.intent.is_group_enabled


def _import_performs():  # noqa: F401
    # isort: off

    # :: Message
    import avilla.qqapi.perform.message.deserialize  # noqa: F401
    import avilla.qqapi.perform.message.serialize  # noqa: F401

    ## :: Action
    import avilla.qqapi.perform.action.user  # noqa: F401
    import avilla.qqapi.perform.action.channel  # noqa: F401
    import avilla.qqapi.perform.action.guild  # noqa: F401
    import avilla.qqapi.perform.action.guild_member  # noqa: F401
    import avilla.qqapi.perform.action.message  # noqa: F401
    import avilla.qqapi.perform.action.role  # noqa: F401

    ## :: Context
    import avilla.qqapi.perform.context  # noqa: F401

    ## :: Event
    import avilla.qqapi.perform.event.activity  # noqa: F401
    import avilla.qqapi.perform.event.audit  # noqa: F401
    import avilla.qqapi.perform.event.metadata  # noqa: F401
    import avilla.qqapi.perform.event.message  # noqa: F401
    import avilla.qqapi.perform.event.relationship  # noqa: F401

    ## :: Query
    import avilla.qqapi.perform.query  # noqa: F401

    ## :: Resource Fetch
    import avilla.qqapi.perform.resource_fetch  # noqa: F401


class QQAPIProtocol(BaseProtocol):
    service: QQAPIService

    def __init__(self):
        self.service = QQAPIService(self)

    _import_performs()
    artifacts = {
        **merge(
            ref("avilla.protocol/qqapi::context"),
            ref("avilla.protocol/qqapi::query"),
            ref("avilla.protocol/qqapi::resource_fetch"),
            ref("avilla.protocol/qqapi::action", "message"),
            ref("avilla.protocol/qqapi::action", "channel"),
            ref("avilla.protocol/qqapi::action", "guild"),
            ref("avilla.protocol/qqapi::action", "guild_member"),
            ref("avilla.protocol/qqapi::action", "role"),
            ref("avilla.protocol/qqapi::action", "user"),
            ref("avilla.protocol/qqapi::message", "deserialize"),
            ref("avilla.protocol/qqapi::message", "serialize"),
            ref("avilla.protocol/qqapi::event", "message"),
            ref("avilla.protocol/qqapi::event", "activity"),
            ref("avilla.protocol/qqapi::event", "relationship"),
            ref("avilla.protocol/qqapi::event", "metadata"),
            ref("avilla.protocol/qqapi::event", "audit"),
        ),
    }

    def ensure(self, avilla: Avilla):
        self.avilla = avilla

        avilla.launch_manager.add_component(self.service)

    def configure(self, config: QQAPIConfig):
        self.service.connections.append(QQAPIWsClientNetworking(self, config))
        return self



from __future__ import annotations

from typing import Any, Literal

from graia.amnesia.message import Element, MessageChain

from avilla.core.event import AvillaEvent
from avilla.core.ryanvk.collector.application import ApplicationCollector
from avilla.core.ryanvk.overload.target import TargetOverload
from avilla.core.selector import Selector
from avilla.standard.core.application.event import AvillaLifecycleEvent
from graia.ryanvk import Fn, PredicateOverload, SimpleOverload, TypeOverload

from .utils import handle_text


class QQAPICapability((m := ApplicationCollector())._):
    @Fn.complex({SimpleOverload(): ["event_type"]})
    async def event_callback(self, event_type: str, raw_event: dict) -> AvillaEvent | AvillaLifecycleEvent | None:
        ...

    @Fn.complex({PredicateOverload(lambda _, raw: raw["type"]): ["raw_element"]})
    async def deserialize_element(self, raw_element: dict) -> Element:
        ...

    @Fn.complex({TypeOverload(): ["element"]})
    async def serialize_element(self, element: Any) -> str | tuple[str, Any]:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def create_dms(self, target: Selector) -> Selector:
        """主动创建私聊会话，返回临时的 Guild"""
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def post_file(
        self,
        target: Selector,
        file_type: Literal[1, 2, 3, 4],
        url: str | None = None,
        srv_send_msg: bool = True,
        file_data: str | bytes | None = None,
    ) -> dict:
        """上传文件，返回文件信息"""
        ...

    async def deserialize(self, event: dict):
        elements = []

        if message_reference := event.get("message_reference"):
            elements.append(await self.deserialize_element({"type": "message_reference", **message_reference}))
        if event.get("mention_everyone", False):
            elements.append(await self.deserialize_element({"type": "mention_everyone"}))
        if "content" in event:
            for i in handle_text(event["content"]):
                elements.append(await self.deserialize_element(i))
        if attachments := event.get("attachments"):
            for i in attachments:
                elements.append(await self.deserialize_element({"type": "attachment", **i}))
        if embeds := event.get("embeds"):
            for i in embeds:
                elements.append(await self.deserialize_element({"type": "embed", **i}))
        if ark := event.get("ark"):
            elements.append(await self.deserialize_element({"type": "ark", **ark}))

        return MessageChain(elements)

    async def serialize(self, message: MessageChain):
        res = {}
        content = ""

        for element in message:
            elem = await self.serialize_element(element)
            if isinstance(elem, str):
                content += elem
            else:
                res[elem[0]] = elem[1]
        if content:
            res["content"] = content
        return res

    async def handle_event(self, etype: str, event: dict):
        maybe_event = await self.event_callback(etype, event)

        if maybe_event is not None:
            self.avilla.broadcast.postEvent(maybe_event)



from __future__ import annotations

from datetime import datetime

from loguru import logger

from avilla.core.context import Context
from avilla.core.elements import Notice, Text
from avilla.core.message import Message
from avilla.core.selector import Selector
from avilla.qqapi.capability import QQAPICapability
from avilla.qqapi.collector.connection import ConnectionCollector
from avilla.qqapi.element import Reference
from avilla.standard.core.message import MessageReceived, MessageRevoked
from avilla.standard.core.profile import Avatar, Nick, Summary


class QQAPIEventMessagePerform((m := ConnectionCollector())._):
    m.namespace = "avilla.protocol/qqapi::event"
    m.identify = "message"

    @m.entity(QQAPICapability.event_callback, event_type="at_message_create")
    @m.entity(QQAPICapability.event_callback, event_type="message_create")
    async def at_message(self, event_type: str, raw_event: dict):
        # TODO: put the author.bot metadata
        account_route = Selector().land("qq").account(self.connection.account_id)
        info = self.protocol.avilla.accounts.get(account_route)
        if info is None:
            logger.warning(f"Unknown account {self.connection.account_id} received message {raw_event}")
            return
        account = info.account
        guild = Selector().land("qq").guild(raw_event["guild_id"])
        channel = guild.channel(raw_event["channel_id"])
        author = channel.member(raw_event["author"]["id"])
        context = Context(
            account,
            author,
            channel,
            channel,
            channel.member(account_route["account"]),
        )
        message = await QQAPICapability(account.staff.ext({"context": context})).deserialize(raw_event)
        reply = None
        if i := message.get(Reference):
            reply = i[0].message
            message = message.exclude(Reference)
        if event_type == "at_message_create" and message.content and isinstance(message.content[1], Text):
            text = message.content[1].text.lstrip()  # type: ignore
            if not text:
                message.content.pop(1)
            else:
                message.content[1] = Text(text)
        msg = Message(
            id=raw_event["id"],
            scene=channel,
            sender=author,
            content=message,
            time=datetime.fromisoformat(raw_event["timestamp"]),
            reply=reply,
        )
        context.cache[context.scene.display_without_land] = raw_event["id"]
        context._collect_metadatas(
            author,
            Nick(
                raw_event["author"]["username"], raw_event["member"].get("nick", raw_event["author"]["username"]), None
            ),
            Summary(raw_event["author"]["username"], "channel member"),
            Avatar(raw_event["author"]["avatar"]),
        )
        context._collect_metadatas(msg.to_selector(), msg)
        return MessageReceived(context, msg)

    @m.entity(QQAPICapability.event_callback, event_type="group_at_message_create")
    async def group_at_message(self, event_type: ..., raw_event: dict):
        account_route = Selector().land("qq").account(self.connection.account_id)
        info = self.protocol.avilla.accounts.get(account_route)
        if info is None:
            logger.warning(f"Unknown account {self.connection.account_id} received message {raw_event}")
            return
        account = info.account
        if "group_openid" in raw_event:
            group = Selector().land("qq").group(raw_event["group_openid"])
        else:
            group = Selector().land("qq").group(raw_event["group_id"])
        if "member_openid" in raw_event["author"]:
            author = group.member(raw_event["author"]["member_openid"])
        else:
            author = group.member(raw_event["author"]["id"])
        context = Context(
            account,
            author,
            group,
            group,
            group.member(account_route["account"]),
        )
        message = await QQAPICapability(account.staff.ext({"context": context})).deserialize(raw_event)
        reply = None
        # TODO: wait for qqapi upgrade
        # if i := message.get(Reference):
        #     reply = group.message(i[0].message_id)
        #     message = message.exclude(Reference)
        message.content.insert(0, Notice(group.member(account_route["account"])))
        if message.content and isinstance(message.content[1], Text):
            text = message.content[1].text.lstrip()  # type: ignore
            if not text:
                message.content.pop(1)
            else:
                message.content[1] = Text(text)
        msg = Message(
            id=raw_event["id"],
            scene=group,
            sender=author,
            content=message,
            time=datetime.fromisoformat(raw_event["timestamp"]),
            reply=reply,
        )
        context.cache[context.scene.display_without_land] = raw_event["id"]
        context._collect_metadatas(msg.to_selector(), msg)
        return MessageReceived(context, msg)

    @m.entity(QQAPICapability.event_callback, event_type="direct_message_create")
    async def direct_message(self, event_type: ..., raw_event: dict):
        account_route = Selector().land("qq").account(self.connection.account_id)
        info = self.protocol.avilla.accounts.get(account_route)
        if info is None:
            logger.warning(f"Unknown account {self.connection.account_id} received message {raw_event}")
            return
        account = info.account
        guild = Selector().land("qq").guild(raw_event["src_guild_id"])
        author = guild.user(raw_event["author"]["id"])
        context = Context(
            account,
            author,
            account_route,
            author,
            account_route,
        )
        message = await QQAPICapability(account.staff.ext({"context": context})).deserialize(raw_event)
        reply = None
        if i := message.get(Reference):
            reply = i[0].message
            message = message.exclude(Reference)

        msg = Message(
            id=raw_event["id"],
            scene=author,
            sender=author,
            content=message,
            time=datetime.fromisoformat(raw_event["timestamp"]),
            reply=reply,
        )
        context.cache[context.scene.display_without_land] = raw_event["id"]
        context.cache[f"{context.scene.display_without_land}#dms"] = raw_event["guild_id"]
        context._collect_metadatas(
            author,
            Nick(
                raw_event["author"]["username"], raw_event["member"].get("nick", raw_event["author"]["username"]), None
            ),
            Summary(raw_event["author"]["username"], "channel member"),
            Avatar(raw_event["author"]["avatar"]),
        )
        context._collect_metadatas(msg.to_selector(), msg)
        return MessageReceived(context, msg)

    @m.entity(QQAPICapability.event_callback, event_type="c2c_message_create")
    async def c2c_message(self, event_type: ..., raw_event: dict):
        account_route = Selector().land("qq").account(self.connection.account_id)
        info = self.protocol.avilla.accounts.get(account_route)
        if info is None:
            logger.warning(f"Unknown account {self.connection.account_id} received message {raw_event}")
            return
        account = info.account
        if "user_openid" in raw_event["author"]:
            friend = Selector().land("qq").friend(raw_event["author"]["user_openid"])
        else:
            friend = Selector().land("qq").friend(raw_event["author"]["id"])
        context = Context(
            account,
            friend,
            account_route,
            friend,
            account_route,
        )
        message = await QQAPICapability(account.staff.ext({"context": context})).deserialize(raw_event)
        reply = None
        if i := message.get(Reference):
            reply = i[0].message
            message = message.exclude(Reference)
        msg = Message(
            id=raw_event["id"],
            scene=friend,
            sender=friend,
            content=message,
            time=datetime.fromisoformat(raw_event["timestamp"]),
            reply=reply,
        )
        context.cache[context.scene.display_without_land] = raw_event["id"]
        context._collect_metadatas(msg.to_selector(), msg)
        return MessageReceived(context, msg)

    @m.entity(QQAPICapability.event_callback, event_type="self_message_create")
    async def self_at_message(self, event_type: ..., raw_event: dict):
        # TODO: put the author.bot metadata
        account_route = Selector().land("qq").account(self.connection.account_id)
        info = self.protocol.avilla.accounts.get(account_route)
        if info is None:
            logger.warning(f"Unknown account {self.connection.account_id} received message {raw_event}")
            return
        account = info.account
        guild = Selector().land("qq").guild(raw_event["guild_id"])
        channel = guild.channel(raw_event["channel_id"])
        author = channel.member(raw_event["author"]["id"])
        context = Context(
            account,
            author,
            channel,
            channel,
            channel.member(account_route["account"]),
        )
        message = await QQAPICapability(account.staff.ext({"context": context})).deserialize(raw_event)
        reply = None
        if i := message.get(Reference):
            reply = i[0].message
            message = message.exclude(Reference)
        msg = Message(
            id=raw_event["id"],
            scene=channel,
            sender=author,
            content=message,
            time=datetime.fromisoformat(raw_event["timestamp"]),
            reply=reply,
        )
        context._collect_metadatas(msg.to_selector(), msg)
        return MessageReceived(context, msg)

    @m.entity(QQAPICapability.event_callback, event_type="self_direct_message_create")
    async def self_direct_message(self, event_type: ..., raw_event: dict):
        account_route = Selector().land("qq").account(self.connection.account_id)
        info = self.protocol.avilla.accounts.get(account_route)
        if info is None:
            logger.warning(f"Unknown account {self.connection.account_id} received message {raw_event}")
            return
        account = info.account
        guild = Selector().land("qq").guild(raw_event["guild_id"])
        author = guild.user(raw_event["author"]["id"])
        context = Context(
            account,
            author,
            account_route,
            author,
            account_route,
        )
        message = await QQAPICapability(account.staff.ext({"context": context})).deserialize(raw_event)
        reply = None
        if i := message.get(Reference):
            reply = i[0].message
            message = message.exclude(Reference)

        msg = Message(
            id=raw_event["id"],
            scene=author,
            sender=author,
            content=message,
            time=datetime.fromisoformat(raw_event["timestamp"]),
            reply=reply,
        )
        context._collect_metadatas(msg.to_selector(), msg)
        return MessageReceived(context, msg)

    @m.entity(QQAPICapability.event_callback, event_type="message_delete")
    @m.entity(QQAPICapability.event_callback, event_type="public_message_delete")
    async def public_message_delete(self, event_type: ..., raw_event: dict):
        account_route = Selector().land("qq").account(self.connection.account_id)
        info = self.protocol.avilla.accounts.get(account_route)
        if info is None:
            logger.warning(f"Unknown account {self.connection.account_id} received message {raw_event}")
            return
        account = info.account
        guild = Selector().land("qq").guild(raw_event["message"]["guild_id"])
        channel = guild.channel(raw_event["message"]["channel_id"])
        author = channel.member(raw_event["message"]["author"]["id"])
        operator = channel.member(raw_event["op_user"]["id"])
        context = Context(
            account,
            operator,
            author,
            channel,
            channel.member(account_route["account"]),
        )
        context._collect_metadatas(
            author,
            Nick(raw_event["message"]["author"]["username"], raw_event["message"]["author"]["username"], None),
            Summary(raw_event["message"]["author"]["username"], "channel member"),
            Avatar(raw_event["message"]["author"]["avatar"]),
        )
        return MessageRevoked(context, author.message(raw_event["message"]["id"]), operator)

    @m.entity(QQAPICapability.event_callback, event_type="direct_message_delete")
    async def direct_message_delete(self, event_type: ..., raw_event: dict):
        account_route = Selector().land("qq").account(self.connection.account_id)
        info = self.protocol.avilla.accounts.get(account_route)
        if info is None:
            logger.warning(f"Unknown account {self.connection.account_id} received message {raw_event}")
            return
        account = info.account
        guild = Selector().land("qq").guild(raw_event["message"]["guild_id"])
        author = guild.user(raw_event["message"]["author"]["id"])
        operator = guild.user(raw_event["op_user"]["id"])
        context = Context(
            account,
            operator,
            author,
            author,
            account_route,
        )
        context._collect_metadatas(
            author,
            Nick(raw_event["message"]["author"]["username"], raw_event["message"]["author"]["username"], None),
            Summary(raw_event["message"]["author"]["username"], "channel member"),
            Avatar(raw_event["message"]["author"]["avatar"]),
        )
        return MessageRevoked(context, author.message(raw_event["message"]["id"]), operator)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Literal

from aiohttp import ClientResponse

from avilla.qqapi.exception import (
    ActionFailed,
    ApiNotAvailable,
    AuditException,
    RateLimitException,
    UnauthorizedException,
)


async def validate_response(resp: ClientResponse):
    status = resp.status
    if status == 200 or 203 <= status < 300:
        data = await resp.json()
        return data.get("data", data)
    if status in {201, 202}:
        data = await resp.json()
        if data and (audit_id := data.get("data", {}).get("message_audit", {}).get("audit_id")):
            exc = AuditException(audit_id)
        else:
            exc = ActionFailed(status, resp.headers, await resp.text())
    elif status == 401:
        exc = UnauthorizedException(status, resp.headers, await resp.text())
    elif status in {404, 405}:
        exc = ApiNotAvailable(status, resp.headers, await resp.text())
    elif status == 429:
        exc = RateLimitException(status, resp.headers, await resp.text())
    else:
        exc = ActionFailed(status, resp.headers, await resp.text())
    raise exc


CallMethod = Literal["get", "post", "fetch", "update", "multipart"]


class Opcode(int, Enum):
    DISPATCH = 0
    HEARTBEAT = 1
    IDENTIFY = 2
    RESUME = 6
    RECONNECT = 7
    INVALID_SESSION = 9
    HELLO = 10
    HEARTBEAT_ACK = 11


@dataclass
class Payload:
    op: int | Opcode
    d: dict | None = None
    s: int | None = None
    t: str | None = None
    id: str | None = None

    @property
    def opcode(self) -> Opcode:
        return Opcode(self.op)

    @property
    def data(self) -> dict:
        return self.d or {}

    @property
    def sequence(self) -> int | None:
        return self.s

    @property
    def type(self) -> str | None:
        return self.t



from __future__ import annotations

from base64 import b64encode
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Any

from graia.amnesia.builtins.memcache import Memcache, MemcacheService
from graia.amnesia.message import MessageChain
from loguru import logger

from avilla.core import Context, CoreCapability, Message
from avilla.core.exceptions import ActionFailed
from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.core.selector import Selector
from avilla.qqapi.capability import QQAPICapability
from avilla.qqapi.exception import AuditException
from avilla.qqapi.utils import form_data, unescape
from avilla.standard.core.message import (
    MessageReceived,
    MessageRevoke,
    MessageSend,
    MessageSent,
)
from graia.ryanvk import OptionalAccess

if TYPE_CHECKING:
    from avilla.qqapi.account import QQAPIAccount  # noqa
    from avilla.qqapi.protocol import QQAPIProtocol  # noqa


class QQAPIMessageActionPerform((m := AccountCollector["QQAPIProtocol", "QQAPIAccount"]())._):
    m.namespace = "avilla.protocol/qqapi::action"
    m.identify = "message"

    context: OptionalAccess[Context] = OptionalAccess()

    @m.entity(QQAPICapability.post_file, target="land.group")
    async def post_group_file(
        self,
        target: Selector,
        file_type: int,
        url: str | None = None,
        srv_send_msg: bool = True,
        file_data: str | bytes | None = None,
    ) -> dict:
        if isinstance(file_data, bytes):
            file_data = b64encode(file_data).decode()
        result = await self.account.connection.call_http(
            "post",
            f"v2/groups/{target.pattern['group']}/files",
            {
                "file_type": file_type,
                "url": url,
                "srv_send_msg": srv_send_msg,
                "file_data": file_data,
            },
        )
        if result is None:
            raise ActionFailed(f"Failed to post file to {target.pattern['group']}")
        return result

    @m.entity(QQAPICapability.post_file, target="land.friend")
    async def post_friend_file(
        self,
        target: Selector,
        file_type: int,
        url: str | None = None,
        srv_send_msg: bool = True,
        file_data: str | bytes | None = None,
    ) -> dict:
        if isinstance(file_data, bytes):
            file_data = b64encode(file_data).decode()
        result = await self.account.connection.call_http(
            "post",
            f"v2/users/{target.pattern['friend']}/files",
            {
                "file_type": file_type,
                "url": url,
                "srv_send_msg": srv_send_msg,
                "file_data": file_data,
            },
        )
        if result is None:
            raise ActionFailed(f"Failed to post file to {target.pattern['friend']}")
        return result

    @staticmethod
    def _extract_qq_media(msg: dict) -> dict[str, Any]:
        kwargs = {}
        if media := msg.get("media"):
            kwargs["file_type"] = {"image": 1, "video": 2, "audio": 3, "file": 4}.get(media[0], 4)
            kwargs["url"] = media[1]
        elif "file_image" in msg:
            kwargs["file_type"] = 1
            kwargs["file_data"] = msg.pop("file_image")
        elif "file_audio" in msg:
            kwargs["file_type"] = 2
            kwargs["file_data"] = msg.pop("file_audio")
        elif "file_video" in msg:
            kwargs["file_type"] = 3
            kwargs["file_data"] = msg.pop("file_video")
        elif "file_file" in msg:
            kwargs["file_type"] = 4
            kwargs["file_data"] = msg.pop("file_file")
        return kwargs

    @MessageSend.send.collect(m, target="land.guild.channel")
    async def send_channel_msg(
        self,
        target: Selector,
        message: MessageChain,
        *,
        reply: Selector | None = None,
    ) -> Selector:
        context = self.context
        msg = await QQAPICapability(self.account.staff).serialize(message)
        if context and (event_id := context.cache.get(target.display_without_land)):
            msg["msg_id"] = event_id
        if reply:
            msg["msg_id"] = reply.pattern["message"]
            msg["message_reference"] = {"message_id": reply.pattern["message"]}
        if media := msg.get("media"):
            msg[media[0]] = media[1]
        method, data = form_data(msg)
        try:
            result = await self.account.connection.call_http(
                method, f"channels/{target.pattern['channel']}/messages", data
            )
            if result is None:
                raise ActionFailed(f"Failed to send message to {target.pattern['channel']}: {message}")
            event = await QQAPICapability(
                self.account.staff.ext({"connection": self.account.connection})
            ).event_callback("self_message_create", result)
            if TYPE_CHECKING:
                assert isinstance(event, MessageReceived)
            event.context = self.account.get_context(target.member(self.account.route["account"]))
            event.message.scene = target
            event.message.content = message
            event.message.sender = target.member(self.account.route["account"])
            self.protocol.post_event(MessageSent(event.context, event.message, self.account))
            return event.message.to_selector()
        except AuditException as e:
            audit_res = await e.get_audit_result()
            if not audit_res or not audit_res.audit.message:
                raise ActionFailed(f"Failed to send message to {target.pattern['channel']}: {message}") from e
            result = await self.account.connection.call_http(
                "get",
                f"channels/{audit_res.audit.message.pattern['channel']}/messages/{audit_res.audit.message.pattern['message']}",
            )
            if result is None:
                logger.warning(f"Failed to get message from {audit_res.audit.message.pattern['channel']}: {message}")
                return audit_res.audit.message
            event = await QQAPICapability(
                self.account.staff.ext({"connection": self.account.connection})
            ).event_callback("self_message_create", result["message"])
            if TYPE_CHECKING:
                assert isinstance(event, MessageReceived)
            event.context = self.account.get_context(target.member(self.account.route["account"]))
            event.message.scene = target
            event.message.sender = target.member(self.account.route["account"])
            self.protocol.post_event(MessageSent(event.context, event.message, self.account))
            return event.message.to_selector()

    @MessageSend.send.collect(m, target="land.group")
    async def send_group_msg(
        self,
        target: Selector,
        message: MessageChain,
        *,
        reply: Selector | None = None,
    ) -> Selector:
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        msg = await QQAPICapability(self.account.staff).serialize(message)
        context = self.context
        if context and (event_id := context.cache.get(target.display_without_land)):
            msg["msg_id"] = event_id
        if reply:
            msg["msg_id"] = reply.pattern["message"]
            # TODO: wait for api upgrade
            # msg["message_reference"] = {"message_id": reply.pattern["message"]}
        if msg.get("embed"):
            msg_type = 4
        elif msg.get("ark"):
            msg_type = 3
        elif msg.get("markdown"):
            msg_type = 2
        elif any(
            i in msg
            for i in (
                "media",
                "file_image",
                "file_video",
                "file_audio",
                "file_file",
            )
        ):
            msg_type = 7
        else:
            msg_type = 0
        msg["timestamp"] = int(datetime.now(timezone.utc).timestamp())
        if msg_type == 7:
            result = await self.post_group_file(target, srv_send_msg="msg_id" not in msg, **self._extract_qq_media(msg))
            if "msg_id" not in msg:
                context = self.account.get_context(target)
                msg = Message(
                    result.get("id", "UNKNOWN"),
                    target,
                    target.member(self.account.route["account"]),
                    message,
                    datetime.now(timezone.utc),
                )
                self.protocol.post_event(MessageSent(context, msg, self.account))
                return target.message(result.get("id", "UNKNOWN"))
            msg["media"] = {"file_info": result["file_info"]}
            if "content" not in msg or not msg["content"]:
                msg["content"] = " "
        msg["msg_type"] = msg_type

        if seq := await cache.get(
            f"qqapi/account({self.account.route['account']}):{target}+msg_id:{msg.get('msg_id', '_')}"
        ):
            msg["msg_seq"] = seq
            await cache.set(
                f"qqapi/account({self.account.route['account']}):{target}+msg_id:{msg.get('msg_id', '_')}",
                seq + 1,
                expire=timedelta(minutes=5),
            )
        else:
            msg["msg_seq"] = 1
            await cache.set(
                f"qqapi/account({self.account.route['account']}):{target}+msg_id:{msg.get('msg_id', '_')}",
                2,
                expire=timedelta(minutes=5),
            )
        # TODO: wait for api upgrade
        msg["content"] = unescape(msg["content"])
        method, data = form_data(msg)
        try:
            result = await self.account.connection.call_http(method, f"v2/groups/{target.pattern['group']}/messages", data)
        except ActionFailed:
            msg["msg_seq"] = msg["msg_seq"] + (hash(target.pattern['group']) % 0x7FFFFFF) + int(datetime.now(timezone.utc).timestamp())
            result = await self.account.connection.call_http(method, f"v2/groups/{target.pattern['group']}/messages", data)
        if result is None:
            raise ActionFailed(f"Failed to send message to {target.pattern['group']}: {message}")
        context = self.account.get_context(target)
        msg = Message(
            result.get("id", "UNKNOWN"),
            target,
            target.member(self.account.route["account"]),
            message,
            datetime.now(timezone.utc),
        )
        self.protocol.post_event(MessageSent(context, msg, self.account))
        return target.message(result.get("id", "UNKNOWN"))

    @MessageSend.send.collect(m, target="land.friend")
    async def send_friend_msg(
        self,
        target: Selector,
        message: MessageChain,
        *,
        reply: Selector | None = None,
    ) -> Selector:
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        context = self.context
        msg = await QQAPICapability(self.account.staff).serialize(message)
        if context and (event_id := context.cache.get(target.display_without_land)):
            msg["msg_id"] = event_id
        if reply:
            msg["msg_id"] = reply.pattern["message"]
            # TODO: wait for api upgrade
            # msg["message_reference"] = {"message_id": reply.pattern["message"]}
        if msg.get("embed"):
            msg_type = 4
        elif msg.get("ark"):
            msg_type = 3
        elif msg.get("markdown"):
            msg_type = 2
        elif any(
            i in msg
            for i in (
                "media",
                "file_image",
                "file_video",
                "file_audio",
                "file_file",
            )
        ):
            msg_type = 7
        else:
            msg_type = 0
        msg["timestamp"] = int(datetime.now(timezone.utc).timestamp())
        if msg_type == 7:
            result = await self.post_friend_file(target, srv_send_msg="msg_id" not in msg, **self._extract_qq_media(msg))
            if "msg_id" not in msg:
                context = self.account.get_context(target)
                msg = Message(
                    result.get("id", "UNKNOWN"),
                    target,
                    target.member(self.account.route["account"]),
                    message,
                    datetime.now(timezone.utc),
                )
                self.protocol.post_event(MessageSent(context, msg, self.account))
                return target.message(result.get("id", "UNKNOWN"))
            msg["media"] = {"file_info": result["file_info"]}
            if "content" not in msg or not msg["content"]:
                msg["content"] = " "
        msg["msg_type"] = msg_type
        if seq := await cache.get(
            f"qqapi/account({self.account.route['account']}):{target}+msg_id:{msg.get('msg_id', '_')}"
        ):
            msg["msg_seq"] = seq
            await cache.set(
                f"qqapi/account({self.account.route['account']}):{target}+msg_id:{msg.get('msg_id', '_')}",
                seq + 1,
                expire=timedelta(minutes=5),
            )
        else:
            msg["msg_seq"] = 1
            await cache.set(
                f"qqapi/account({self.account.route['account']}):{target}+msg_id:{msg.get('msg_id', '_')}",
                2,
                expire=timedelta(minutes=5),
            )
        # TODO: wait for api upgrade
        msg["content"] = unescape(msg["content"])
        method, data = form_data(msg)
        try:
            result = await self.account.connection.call_http(method, f"v2/users/{target.pattern['friend']}/messages", data)
        except ActionFailed:
            msg["msg_seq"] = msg["msg_seq"] + (hash(target.pattern['friend']) % 0x7FFFFFF) + int(datetime.now(timezone.utc).timestamp())
            result = await self.account.connection.call_http(method, f"v2/users/{target.pattern['friend']}/messages", data)
        if result is None:
            raise ActionFailed(f"Failed to send message to {target.pattern['friend']}: {message}")
        context = self.account.get_context(target)
        msg = Message(
            result.get("id", "UNKNOWN"),
            target,
            target.member(self.account.route["account"]),
            message,
            datetime.now(timezone.utc),
        )
        self.protocol.post_event(MessageSent(context, msg, self.account))
        return target.message(result.get("id", "UNKNOWN"))

    @QQAPICapability.create_dms.collect(m, target="land.guild.user")
    async def create_dms_user(self, target: Selector) -> Selector:
        result = await self.account.connection.call_http(
            "post",
            "users/@me/dms",
            {
                "recipient_id": target.pattern["user"],
                "source_guild_id": target.pattern["guild"],
            },
        )
        context = self.context
        if context:
            context.cache[f"{context.scene.display_without_land}#dms"] = result["guild_id"]
        return target.into(f"land.guild({result['guild_id']})")

    @QQAPICapability.create_dms.collect(m, target="land.guild.member")
    @QQAPICapability.create_dms.collect(m, target="land.guild.channel.member")
    async def create_dms_member(self, target: Selector) -> Selector:
        result = await self.account.connection.call_http(
            "post",
            "users/@me/dms",
            {
                "recipient_id": target.pattern["member"],
                "source_guild_id": target.pattern["guild"],
            },
        )
        context = self.context
        if context:
            context.cache[f"{context.scene.display_without_land}#dms"] = result["guild_id"]
        return target.into(f"land.guild({result['guild_id']})")

    @MessageSend.send.collect(m, target="land.guild.user")
    async def send_direct_msg(
        self,
        target: Selector,
        message: MessageChain,
        *,
        reply: Selector | None = None,
    ) -> Selector:
        context = self.context
        msg = await QQAPICapability(self.account.staff).serialize(message)
        if not context or not (send_guild_id := context.cache.get(f"{context.scene.display_without_land}#dms")):
            send_guild_id = (await self.create_dms_user(target)).pattern["guild"]
        if context and (event_id := context.cache.get(target.display_without_land)):
            msg["msg_id"] = event_id
        if reply:
            msg["msg_id"] = reply.pattern["message"]
            msg["message_reference"] = {"message_id": reply.pattern["message"]}
        if media := msg.get("media"):
            msg[media[0]] = media[1]
        method, data = form_data(msg)
        result = await self.account.connection.call_http(method, f"dms/{send_guild_id}/messages", data)
        if result is None:
            raise ActionFailed(f"Failed to send message to {target.pattern['user']}: {message}")
        event = await QQAPICapability(self.account.staff.ext({"connection": self.account.connection})).event_callback(
            "self_direct_message_create", result
        )
        if TYPE_CHECKING:
            assert isinstance(event, MessageReceived)
        event.context = self.account.get_context(target)
        event.message.scene = target
        event.message.content = message
        event.message.sender = self.account.route
        self.protocol.post_event(MessageSent(event.context, event.message, self.account))
        return event.message.to_selector()

    @MessageRevoke.revoke.collect(m, target="land.guild.channel.message")
    async def delete_msg(self, target: Selector):
        await self.account.connection.call_http(
            "delete",
            f"channels/{target.pattern['channel']}/messages/{target.pattern['message']}",
            {"hidetip": str(False).lower()},
        )

    @MessageRevoke.revoke.collect(m, target="land.guild.user.message")
    async def delete_direct_msg(self, target: Selector):
        context = self.context
        if not context or not (send_guild_id := context.cache.get(f"{context.scene.display_without_land}#dms")):
            send_guild_id = (await self.create_dms_user(target)).pattern["guild"]
            if context:
                context.cache.pop(f"{context.scene.display_without_land}#dms", None)  # type: ignore
        await self.account.connection.call_http(
            "delete",
            f"dms/{send_guild_id}/messages/{target.pattern['message']}",
            {"hidetip": str(False).lower()},
        )

    @m.entity(CoreCapability.pull, target="land.guild.channel.message", route=Message)
    async def get_group_message(self, target: Selector, route: ...) -> Message:
        result = await self.account.connection.call_http(
            "get",
            f"channels/{target.pattern['channel']}/messages/{target.pattern['message']}",
        )
        if result is None:
            raise RuntimeError(f"Failed to get message from {target.pattern['channel']}: {target}")
        event = await QQAPICapability(self.account.staff.ext({"connection": self.account.connection})).event_callback(
            "message_create", result
        )
        if TYPE_CHECKING:
            assert isinstance(event, MessageReceived)  # noqa
        return event.message



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING

from avilla.core.exceptions import permission_error_message
from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.core.selector import Selector
from avilla.qqapi.const import PRIVILEGE_TRANS
from avilla.qqapi.role import (
    Role,
    RoleCreate,
    RoleDelete,
    RoleEdit,
    RoleMemberCapability,
)
from avilla.standard.core.common import Count
from avilla.standard.core.privilege import MuteCapability, MuteInfo, Privilege
from avilla.standard.core.profile import Summary, SummaryCapability

if TYPE_CHECKING:
    from avilla.qqapi.account import QQAPIAccount  # noqa
    from avilla.qqapi.protocol import QQAPIProtocol  # noqa


class QQAPIRoleActionPerform((m := AccountCollector["QQAPIProtocol", "QQAPIAccount"]())._):
    m.namespace = "avilla.protocol/qqapi::action"
    m.identify = "role"

    @m.pull("land.guild.role", Role)
    async def get_role(self, target: Selector, route: ...) -> Role:
        result = await self.account.connection.call_http("get", f"guilds/{target.pattern['guild']}/roles", {})
        for role in result["roles"]:
            if role["id"] == target.pattern["role"]:
                return Role(role["id"], target.into("land.guild"), role["name"], bool(role["hoist"]), role["color"])
        raise ValueError(f"Role {target.pattern['role']} not found")

    @m.pull("land.guild.role", Summary)
    async def get_role_summary(self, target: Selector, route: ...) -> Summary:
        return Summary(
            (await self.get_role(target, Role)).name,
            "name of role",
        )

    @m.pull("land.guild.role", Role >> Summary)
    async def get_role_summary1(self, target: Selector, route: ...) -> Summary:
        return (await self.get_role_summary(target, Summary)).infers(Role >> Summary)

    @m.pull("land.guild.role", Count)
    async def get_role_count(self, target: Selector, route: ...) -> Count:
        result = await self.account.connection.call_http("get", f"guilds/{target.pattern['guild']}/roles", {})
        for role in result["roles"]:
            if role["id"] == target.pattern["role"]:
                return Count(role["number"], role["member_limit"])
        raise ValueError(f"Role {target.pattern['role']} not found")

    @m.pull("land.guild.role", Role >> Count)
    async def get_role_count1(self, target: Selector, route: ...) -> Count:
        return (await self.get_role_count(target, Count)).infers(Role >> Count)

    @m.pull("land.guild.role", Privilege)
    async def get_privilege(self, target: Selector, route: ...) -> Privilege:
        self_info = await self.account.connection.call_http(
            "get", f"guilds/{target.pattern['guild']}/members/{self.account.route['account']}", {}
        )
        effective = bool({"2", "4", "5"} & set(self_info["roles"]))
        apis = await self.account.connection.call_http("get", f"guilds/{target.pattern['guild']}/api_permission", {})
        for api in apis["apis"]:
            if api["path"] == "/channels/{channel_id}/roles/{role_id}/permissions" and api["method"] == "GET":
                result = await self.account.connection.call_http(
                    "get", f"channels/{target.pattern['channel']}/roles/{target.pattern['role']}/permissions", {}
                )
                return Privilege(
                    int(result["permissions"]) & 2 == 2,
                    effective,
                )
        raise PermissionError(permission_error_message(f"get_permission@{target.path}", "read", ["manage"]))

    @m.pull("land.guild.role", Role >> Privilege)
    async def get_privilege1(self, target: Selector, route: ...) -> Privilege:
        return (await self.get_privilege(target, Privilege)).infers(Role >> Privilege)

    @m.pull("land.guild.role", Privilege >> Summary)
    async def get_privilege_summary(self, target: Selector, route: ...) -> Summary:
        apis = await self.account.connection.call_http("get", f"guilds/{target.pattern['guild']}/api_permission", {})
        for api in apis["apis"]:
            if api["path"] == "/channels/{channel_id}/roles/{role_id}/permissions" and api["method"] == "GET":
                result = await self.account.connection.call_http(
                    "get", f"channels/{target.pattern['channel']}/roles/{target.pattern['role']}/permissions", {}
                )
                return Summary(
                    PRIVILEGE_TRANS[int(result["permissions"])],
                    "user's permissions in channel",
                )
        raise PermissionError(permission_error_message(f"get_permission@{target.path}", "read", ["manage"]))

    @m.pull("land.guild.role", Role >> Privilege >> Summary)
    async def get_privilege_summary1(self, target: Selector, route: ...) -> Summary:
        return (await self.get_privilege_summary(target, Summary)).infers(Role >> Privilege >> Summary)

    @m.pull("land.guild.role", MuteInfo)
    async def get_mute_info(self, target: Selector, route: ...) -> MuteInfo:
        apis = await self.account.connection.call_http("get", f"guilds/{target.pattern['guild']}/api_permission", {})
        for api in apis["apis"]:
            if api["path"] == "/channels/{channel_id}/roles/{role_id}/permissions" and api["method"] == "GET":
                result = await self.account.connection.call_http(
                    "get", f"channels/{target.pattern['channel']}/roles/{target.pattern['role']}/permissions", {}
                )
                return MuteInfo(
                    int(result["permissions"]) & 4 == 4,
                    timedelta(seconds=0),
                )
        raise PermissionError(permission_error_message(f"get_permission@{target.path}", "read", ["manage"]))

    @MuteCapability.mute.collect(m, target="land.guild.role")
    async def mute(self, target: Selector, duration: timedelta) -> None:
        if not (await self.get_privilege(target, Privilege)).effective:
            raise PermissionError(permission_error_message(f"set_permission@{target.path}", "read", ["manage"]))
        await self.account.connection.call_http(
            "put",
            f"channels/{target.pattern['channel']}/roles/{target.pattern['role']}/permissions",
            {"add": 0, "remove": 4},
        )

    @MuteCapability.unmute.collect(m, target="land.guild.role")
    async def unmute(self, target: Selector) -> None:
        if not (await self.get_privilege(target, Privilege)).effective:
            raise PermissionError(permission_error_message(f"set_permission@{target.path}", "read", ["manage"]))
        await self.account.connection.call_http(
            "put",
            f"channels/{target.pattern['channel']}/roles/{target.pattern['role']}/permissions",
            {"add": 4, "remove": 0},
        )

    @RoleCreate.create.collect(m, target="land.guild")
    async def create_role(
        self, target: Selector, name: str, hoist: bool | None = None, color: int | None = None
    ) -> Selector:
        self_info = await self.account.connection.call_http(
            "get", f"guilds/{target.pattern['guild']}/members/{self.account.route['account']}", {}
        )
        effective = bool({"2", "4", "5"} & set(self_info["roles"]))
        if not effective:
            raise PermissionError(permission_error_message(f"create_role@{target.path}", "read", ["manage"]))
        data: dict = {"name": name}
        if hoist is not None:
            data["hoist"] = hoist
        if color is not None:
            data["color"] = color
        result = await self.account.connection.call_http("post", f"guilds/{target.pattern['guild']}/roles", data)
        return target.role(result["role_id"])

    @RoleDelete.delete.collect(m, target="land.guild.role")
    async def delete_role(self, target: Selector) -> None:
        self_info = await self.account.connection.call_http(
            "get", f"guilds/{target.pattern['guild']}/members/{self.account.route['account']}", {}
        )
        effective = bool({"2", "4", "5"} & set(self_info["roles"]))
        if not effective:
            raise PermissionError(permission_error_message(f"delete_role@{target.path}", "read", ["manage"]))
        await self.account.connection.call_http(
            "delete", f"guilds/{target.pattern['guild']}/roles/{target.pattern['role']}", {}
        )

    @SummaryCapability.set_name.collect(m, target="land.guild.role", route=Summary)
    async def set_name(self, target: Selector, route: ..., name: str) -> None:
        self_info = await self.account.connection.call_http(
            "get", f"guilds/{target.pattern['guild']}/members/{self.account.route['account']}", {}
        )
        effective = bool({"2", "4", "5"} & set(self_info["roles"]))
        if not effective:
            raise PermissionError(permission_error_message(f"set_name@{target.path}", "read", ["manage"]))
        await self.account.connection.call_http(
            "patch", f"guilds/{target.pattern['guild']}/roles/{target.pattern['role']}", {"name": name}
        )

    @RoleEdit.edit.collect(m, target="land.guild.role")
    async def edit_role(
        self, target: Selector, name: str | None = None, hoist: bool | None = None, color: int | None = None
    ) -> None:
        self_info = await self.account.connection.call_http(
            "get", f"guilds/{target.pattern['guild']}/members/{self.account.route['account']}", {}
        )
        effective = bool({"2", "4", "5"} & set(self_info["roles"]))
        if not effective:
            raise PermissionError(permission_error_message(f"edit_role@{target.path}", "read", ["manage"]))
        data = {}
        if name is not None:
            data["name"] = name
        if hoist is not None:
            data["hoist"] = hoist
        if color is not None:
            data["color"] = color
        await self.account.connection.call_http(
            "patch", f"guilds/{target.pattern['guild']}/roles/{target.pattern['role']}", data
        )

    @RoleMemberCapability.add.collect(m, target="land.guild.user")
    async def add_role_user(self, target: Selector, member: Selector) -> None:
        role_id = target.pattern["role"]
        if role_id == "5":
            raise ValueError("missing target: channel")
        await self.account.connection.call_http(
            "put", f"guilds/{target.pattern['guild']}/members/{member.pattern['user']}/roles/{role_id}", {}
        )

    @RoleMemberCapability.add.collect(m, target="land.guild.member")
    async def add_role_member(self, target: Selector, member: Selector) -> None:
        role_id = target.pattern["role"]
        if role_id == "5":
            raise ValueError("missing target: channel")
        await self.account.connection.call_http(
            "put", f"guilds/{target.pattern['guild']}/members/{member.pattern['member']}/roles/{role_id}", {}
        )

    @RoleMemberCapability.add.collect(m, target="land.guild.channel.member")
    async def add_role_member1(self, target: Selector, member: Selector) -> None:
        role_id = target.pattern["role"]
        await self.account.connection.call_http(
            "put",
            f"guilds/{target.pattern['guild']}/members/{member.pattern['member']}/roles/{role_id}",
            {"channel": {"id": target.pattern["channel"]}} if role_id == "5" else {},
        )

    @RoleMemberCapability.remove.collect(m, target="land.guild.user")
    async def remove_role_user(self, target: Selector, member: Selector) -> None:
        role_id = target.pattern["role"]
        if role_id == "5":
            raise ValueError("missing target: channel")
        await self.account.connection.call_http(
            "delete", f"guilds/{target.pattern['guild']}/members/{member.pattern['user']}/roles/{role_id}", {}
        )

    @RoleMemberCapability.remove.collect(m, target="land.guild.member")
    async def remove_role_member(self, target: Selector, member: Selector) -> None:
        role_id = target.pattern["role"]
        if role_id == "5":
            raise ValueError("missing target: channel")
        await self.account.connection.call_http(
            "delete", f"guilds/{target.pattern['guild']}/members/{member.pattern['member']}/roles/{role_id}", {}
        )

    @RoleMemberCapability.remove.collect(m, target="land.guild.channel.member")
    async def remove_role_member1(self, target: Selector, member: Selector) -> None:
        role_id = target.pattern["role"]
        await self.account.connection.call_http(
            "delete",
            f"guilds/{target.pattern['guild']}/members/{member.pattern['member']}/roles/{role_id}",
            {"channel": {"id": target.pattern["channel"]}} if role_id == "5" else {},
        )



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, TypeVar

from avilla.core.ryanvk.collector.base import AvillaBaseCollector
from graia.ryanvk import Access, BasePerform

if TYPE_CHECKING:
    from avilla.qqapi.connection.ws_client import QQAPIWsClientNetworking
    from avilla.qqapi.protocol import QQAPIProtocol


T = TypeVar("T")
T1 = TypeVar("T1")


class ConnectionBasedPerformTemplate(BasePerform, native=True):
    __collector__: ClassVar[ConnectionCollector]

    protocol: Access[QQAPIProtocol] = Access()
    connection: Access[QQAPIWsClientNetworking] = Access()


class ConnectionCollector(AvillaBaseCollector):
    post_applying: bool = False

    @property
    def _(self):
        upper = super()._

        class PerformTemplate(
            ConnectionBasedPerformTemplate,
            upper,
            native=True,
        ):
            ...

        return PerformTemplate
